try:
    import msgpack

    _HAVE_MSGPACK = True
except ModuleNotFoundError:
    _HAVE_MSGPACK = False

from typing import Any

from communica.utils import ByteSeq
from communica.exceptions import FeatureNotAvailable

from .base import BaseSerializer


class MsgpackSerializer(BaseSerializer):
    """
    This serializer just passes data to json dump function.
    """
    __slots__ = ()

    def __init__(self) -> None:
        if not _HAVE_MSGPACK:
            raise FeatureNotAvailable(
                'MsgpackSerializer requires msgpack library. '
                'Install communica with [msgpack] extra.'
            )

    def dump(self, data: Any) -> bytes:
        dumped = msgpack.packb(data)
        assert isinstance(dumped, bytes)
        return dumped

    def load(self, raw_data: ByteSeq) -> Any:
        return msgpack.unpackb(raw_data)



from .base import BaseSerializer
from .json import JsonSerializer
from .adaptix import AdaptixSerializer
from .msgpack import MsgpackSerializer


default_serializer = JsonSerializer()


__all__ = (
    'BaseSerializer',
    'JsonSerializer',
    'AdaptixSerializer',
    'MsgpackSerializer',
)



import random
import asyncio


class BackoffDelayer:
    __slots__ = ('start_delay', 'max_delay', 'factor', 'jitter', '_delay')

    def __init__(
            self,
            start_delay: float,
            max_delay: float,
            factor: float,
            jitter: float
    ) -> None:
        assert factor > 1, 'factor less or equal 1'
        assert max_delay > start_delay, 'max delay less or equal to start delay'
        self.factor = factor
        self.jitter = jitter

        self._delay = start_delay
        self.max_delay = max_delay
        self.start_delay = start_delay

    def reset(self):
        self._delay = self.start_delay

    async def wait(self):
        await asyncio.sleep(self._delay)
        new_delay = random.gauss(
            min(self.max_delay, (self._delay * self.factor)),
            self.jitter
        )
        if new_delay > self._delay:
            self._delay = new_delay


def default_backoff_delayer():
    return BackoffDelayer(
        start_delay=1,
        max_delay=30,
        factor=2,
        jitter=0.5
    )



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Callable, Protocol
from inspect import iscoroutinefunction
from functools import wraps, partial
from concurrent.futures import ThreadPoolExecutor

from typing_extensions import Self, TypeAlias

from communica.utils import HasLoopMixin, BackoffDelayer, iscallable
from communica.utils.delayer import default_backoff_delayer
from communica.connectors.base import BaseConnector


class BaseEntity(ABC, HasLoopMixin):
    __slots__ = ('connector', '_closed', '_stop_event')

    connector: BaseConnector

    def __init__(self, connector: BaseConnector) -> None:
        self.connector = connector

    def __repr__(self) -> str:
        return f'{type(self).__name__}()'

    async def run(self):
        """Run until .stop() is called."""
        if hasattr(self, '_stop_event'):
            raise RuntimeError('.run() must be called only once on same object')

        self._closed = asyncio.get_event_loop().create_future()
        self._stop_event = asyncio.Event()
        try:
            async with self:
                print(f'Running {self!r} on {self.connector.repr_address()}')
                await self._stop_event.wait()
        finally:
            del(self._stop_event)
            self._closed.set_result(None)

    def stop(self):
        """If .run() was called, it returns, otherwise this method does nothing"""
        if hasattr(self, '_stop_event'):
            self._stop_event.set()

    async def wait_stop(self):
        """Call .stop() and wait until closed"""
        if hasattr(self, '_stop_event'):
            self.stop()
            await self._closed

    @abstractmethod
    async def init(self) -> Self:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    async def __aenter__(self) -> Self:
        return await self.init()

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close()


class BaseClient(BaseEntity):
    __slots__ = ()

    @abstractmethod
    async def init(
        self,
        timeout: 'int | None' = 0,
        delayer_factory: Callable[[], BackoffDelayer] = default_backoff_delayer,
    ) -> Self:
        """
        Establish connection with server.

        On timeout cancels connect task.

        Args:
            timeout: Time in seconds.
              If omitted or zero, this method will block until connect succeed.
              If None, connection initiation will start and any
              .request() or .throw() calls will block until connect succeed.
            delayer_factory: function which returns
              communica.utils.delayer.BackoffDelayer instance.
              By default communica.utils.delayer.default_backoff_delayer is used.

        Raises:
            asyncio.TimeoutError
        """
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Close connection with server."""
        raise NotImplementedError


class BaseServer(BaseEntity):
    __slots__ = ()

    @abstractmethod
    async def init(self) -> Self:
        """Create server and start accepting connections"""
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        """Stop serving and cancel all handler tasks"""
        raise NotImplementedError


HandlerType: TypeAlias = 'SyncHandlerType | AsyncHandlerType | RequestHandler'


class SyncHandlerType(Protocol):
    def __call__(self, data: Any) -> Any: ...


class AsyncHandlerType(Protocol):
    async def __call__(self, data: Any) -> Any: ...


class RequestHandler:
    __slots__ = ('_repr', 'is_async', 'endpoint')

    is_async: bool
    endpoint: 'SyncHandlerType | AsyncHandlerType'

    def __repr__(self) -> str:
        try:
            return self._repr
        except AttributeError:
            name = getattr(self.endpoint, '__qualname__',
                           getattr(self.endpoint, '__name__', 'UNKNOWN'))
            htype = 'async' if self.is_async else 'sync'
            self._repr = f'<RequestHandler {htype} endpoint={name}>'
            return self._repr

    def __init__(self, endpoint: 'SyncHandlerType | AsyncHandlerType') -> None:
        if not iscallable(endpoint):
            raise TypeError('Request handler must be function, '
                            'coroutine function or method')
        self.is_async = iscoroutinefunction(endpoint)
        self.endpoint = endpoint


def threaded_handler(executor: 'ThreadPoolExecutor | None' = None):
    """
    Returns decorator, which returns RequestHandler, which runs
    decorated function in separate thread.
    """
    def decorator(handler: SyncHandlerType):
        # TODO: test with auto serializer
        @wraps(handler)
        async def wrapper(*args, **kwargs):
            loop = asyncio.get_running_loop()
            func = partial(handler, *args, **kwargs)
            return await loop.run_in_executor(executor, func)
        return RequestHandler(handler)
    return decorator
